# coding=utf-8
# Copyright 2025 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""In-Memory implementation of HopliteDBInterface for testing."""

import collections
import dataclasses
from typing import Any, Sequence

from ml_collections import config_dict
import numpy as np
from perch_hoplite.db import interface


@dataclasses.dataclass
class InMemoryGraphSearchDB(interface.HopliteDBInterface):
  """Graph search database backed by NumPy arrays.

  We add embeddings to a fixed-sized array, keeping track of which indices are
  occupied. This avoids copies and resizing of the embeddings table.
  """

  embedding_dim: int
  max_size: int
  embeddings: np.ndarray
  embedding_sources: dict[int, interface.EmbeddingSource] = dataclasses.field(
      default_factory=dict
  )
  embedding_ids: set[int] = dataclasses.field(default_factory=set)
  embedding_dtype: type[Any] = np.float16
  labels: dict[int, list[interface.Label]] = dataclasses.field(
      default_factory=lambda: collections.defaultdict(list)
  )
  kv_store: dict[str, config_dict.ConfigDict] = dataclasses.field(
      default_factory=dict
  )

  @classmethod
  def create(cls, **kwargs):
    """Connect to and, if needed, initialize the database."""
    if 'embeddings' in kwargs:
      db = cls(**kwargs)
      return db

    embeddings = np.zeros(
        [kwargs['max_size'], kwargs['embedding_dim']],
        dtype=kwargs.get('embedding_dtype', np.float16),
    )
    db = cls(embeddings=embeddings, **kwargs)
    return db

  def thread_split(self) -> interface.HopliteDBInterface:
    """Return a readable instance of the database."""
    # Since numpy arrays are in shared memory, we can reuse the same object.
    return self

  def count_embeddings(self) -> int:
    """Return a count of all embeddings in the database."""
    return len(self.embedding_ids)

  def insert_metadata(self, key: str, value: config_dict.ConfigDict) -> None:
    self.kv_store[key] = value

  def get_metadata(self, key: str | None) -> config_dict.ConfigDict:
    if key is None:
      return config_dict.ConfigDict(self.kv_store)
    return self.kv_store[key]

  def get_dataset_names(self) -> Sequence[str]:
    """Get all dataset names in the database."""
    ds_names = set()
    for source in self.embedding_sources.values():
      ds_names.add(source.dataset_name)
    return tuple(ds_names)

  def embedding_dimension(self) -> int:
    return self.embedding_dim

  def insert_embedding(
      self, embedding: np.ndarray, source: interface.EmbeddingSource
  ) -> int:
    """Add an embedding to the database."""
    if len(self.embedding_ids) >= self.max_size:
      # TODO(tomdenton): Automatically resize instead of throwing an error.
      raise ValueError('No more space in predefined memory.')

    if len(self.embedding_ids) not in self.embedding_ids:
      idx = len(self.embedding_ids)
    else:
      # find an unused index...
      idx = None
      for candidate in range(self.max_size):
        if candidate not in self.embedding_ids:
          idx = candidate
      if not idx:
        raise ValueError('Could not find an unused index.')

    self.embedding_ids.add(idx)
    self.embeddings[idx] = embedding[np.newaxis, :]
    self.embedding_sources[idx] = source
    return idx

  def get_embedding(self, embedding_id: int) -> np.ndarray:
    """Retrieve an embedding from the database."""
    return self.embeddings[embedding_id]

  def get_embedding_source(
      self, embedding_id: int
  ) -> interface.EmbeddingSource:
    return self.embedding_sources[embedding_id]

  def get_embeddings(
      self, embedding_ids: np.ndarray
  ) -> tuple[np.ndarray, np.ndarray]:
    return embedding_ids, self.embeddings[embedding_ids]

  def get_embedding_ids(self) -> np.ndarray:
    """Get all embedding IDs in the database."""
    return np.array(tuple(int(id_) for id_ in self.embedding_ids))

  def get_one_embedding_id(self) -> int:
    return next(iter(self.embedding_ids))

  def get_embeddings_by_source(
      self,
      dataset_name: str,
      source_id: str | None,
      offsets: np.ndarray | None = None,
  ) -> np.ndarray:
    found_idxes = set()
    for idx, embedding_source in self.embedding_sources.items():
      if dataset_name and dataset_name != embedding_source.dataset_name:
        continue
      if source_id is not None and source_id != embedding_source.source_id:
        continue
      if offsets is not None and not np.array_equal(
          offsets, embedding_source.offsets
      ):
        continue
      found_idxes.add(idx)
    return np.array(tuple(found_idxes), np.int64)

  def commit(self) -> None:
    """Commit any pending transactions to the database."""
    pass

  def insert_label(
      self, label: interface.Label, skip_duplicates: bool = False
  ) -> bool:
    if label.type is None:
      raise ValueError('label type must be set')
    if label.provenance is None:
      raise ValueError('label source must be set')
    if skip_duplicates and label in self.get_labels(label.embedding_id):
      return False

    self.labels[label.embedding_id].append(label)
    return True

  def get_embeddings_by_label(
      self,
      label: str,
      label_type: interface.LabelType | None = interface.LabelType.POSITIVE,
      provenance: str | None = None,
  ) -> np.ndarray:
    found_idxes = set()
    for idx, emb_labels in self.labels.items():
      for emb_label in emb_labels:
        if emb_label.label != label:
          continue
        if label_type is not None and emb_label.type.value != label_type.value:
          continue
        if provenance is not None and emb_label.provenance != provenance:
          continue
        found_idxes.add(idx)
    return np.array(tuple(found_idxes), np.int64)

  def get_labels(self, embedding_id: int) -> Sequence[interface.Label]:
    return self.labels[embedding_id]

  def get_classes(self) -> Sequence[str]:
    label_set = set()
    for labels in self.labels.values():
      for l in labels:
        label_set.add(l.label)
    return tuple(sorted(label_set))

  def get_class_counts(
      self, label_type: interface.LabelType = interface.LabelType.POSITIVE
  ) -> dict[str, int]:
    class_counts = collections.defaultdict(int)
    for labels in self.labels.values():
      counted_labels = set()
      for l in labels:
        # Avoid double-counting the same label on the same embedding because of
        # different provenances.
        if l.label in counted_labels:
          continue
        if l.type.value == label_type.value:
          class_counts[l.label] += 1
          counted_labels.add(l.label)
        else:
          # Creates a key in the dict for all labels, even if they have no
          # matching type counts.
          class_counts[l.label] += 0
    return class_counts



# coding=utf-8
# Copyright 2025 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for Hoplite databases."""

import shutil
import tempfile

import numpy as np
from perch_hoplite.db import brutalism
from perch_hoplite.db import interface
from perch_hoplite.db.tests import test_utils

from absl.testing import absltest
from absl.testing import parameterized

EMBEDDING_SIZE = 128
DB_TYPES = ('in_mem', 'sqlite_usearch')
DB_TYPE_NAMED_PAIRS = (
    ('in_mem-sqlite_usearch', 'in_mem', 'sqlite_usearch'),
)
PERSISTENT_DB_TYPES = ('sqlite_usearch',)


class HopliteTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.tempdir = tempfile.mkdtemp()

  def tearDown(self):
    super().tearDown()
    shutil.rmtree(self.tempdir)

  @parameterized.product(
      db_type=DB_TYPES,
      thread_split=(True, False),
  )
  def test_graph_db_interface(self, db_type, thread_split):
    rng = np.random.default_rng(42)
    db = test_utils.make_db(self.tempdir, db_type, 1000, rng, EMBEDDING_SIZE)
    if thread_split:
      db = db.thread_split()

    # Run all methods in the interface...
    self.assertEqual(db.count_embeddings(), 1000)
    idxes = db.get_embedding_ids()

    one_idx = db.get_one_embedding_id()
    self.assertIn(one_idx, idxes)

    # Check the metadata.
    got_md = db.get_metadata('db_config')
    self.assertEqual(got_md.embedding_dim, EMBEDDING_SIZE)

    with self.subTest('test_embedding_sources'):
      source = db.get_embedding_source(idxes[1])
      # The embeddings are given one of three randomly selected dataset names.
      embs = db.get_embeddings_by_source(source.dataset_name, None, None)
      self.assertGreater(embs.shape[0], db.count_embeddings() / 6)
      # For an unknown dataset name, we should get no embeddings.
      embs = db.get_embeddings_by_source('fake_name', None, None)
      self.assertEqual(embs.shape[0], 0)
      # Source ids are approximately unique.
      embs = db.get_embeddings_by_source(
          source.dataset_name, source.source_id, None
      )
      self.assertLen(embs, 1)
      # For an unknown source id, we should get no embeddings.
      embs = db.get_embeddings_by_source(source.dataset_name, 'fake_id', None)
      self.assertEqual(embs.shape[0], 0)

      sources = db.get_embedding_sources(idxes[:3])
      self.assertLen(sources, 3)
    db.commit()

  @parameterized.product(db_type=PERSISTENT_DB_TYPES)
  def test_persistence(self, db_type):
    rng = np.random.default_rng(42)
    db = test_utils.make_db(self.tempdir, db_type, 1000, rng, EMBEDDING_SIZE)
    one_emb = np.random.normal(size=(EMBEDDING_SIZE,), loc=0, scale=0.05)
    one_emb_id = db.insert_embedding(
        one_emb, source=interface.EmbeddingSource('q', 'x', np.array([5.0]))
    )
    self.assertEqual(db.get_embedding_ids().shape[0], 1001)
    db.commit()

    got_emb = db.get_embedding(one_emb_id)
    np.testing.assert_equal(got_emb, np.float16(one_emb))

    # "Making" the persistent DB without adding any new embeddings gives us a
    # view of the saved DB.
    test_db = test_utils.make_db(self.tempdir, db_type, 0, rng, EMBEDDING_SIZE)
    self.assertIn(one_emb_id, test_db.get_embedding_ids())
    # Check that the embeddings are the same in the two DB's.
    for idx in db.get_embedding_ids():
      emb = db.get_embedding(idx)
      test_emb = test_db.get_embedding(idx)
      np.testing.assert_equal(emb, test_emb)

  @parameterized.product(db_type=DB_TYPES)
  def test_labels_db_interface(self, db_type):
    rng = np.random.default_rng(42)
    db = test_utils.make_db(self.tempdir, db_type, 1000, rng, EMBEDDING_SIZE)
    ids = db.get_embedding_ids()
    db.insert_label(
        interface.Label(ids[0], 'hawgoo', interface.LabelType.POSITIVE, 'human')
    )
    db.insert_label(
        interface.Label(
            ids[0], 'hawgoo', interface.LabelType.POSITIVE, 'machine'
        )
    )
    db.insert_label(
        interface.Label(
            ids[1], 'hawgoo', interface.LabelType.POSITIVE, 'machine'
        )
    )
    db.insert_label(
        interface.Label(
            ids[0], 'rewbla', interface.LabelType.NEGATIVE, 'machine'
        )
    )

    with self.subTest('get_embeddings_by_label'):
      # When both label_type and provenance are unspecified, we should get all
      # unique IDs with the target label. Id's 0 and 1 both have some kind of
      # 'hawgoo' label.
      got = db.get_embeddings_by_label('hawgoo', None, None)
      self.assertSequenceEqual(sorted(got), sorted([ids[0], ids[1]]))

    with self.subTest('get_embeddings_by_label_type'):
      # Now we should get the ID's for all POSITIVE 'hawgoo' labels, regardless
      # of provenance.
      got = db.get_embeddings_by_label(
          'hawgoo', interface.LabelType.POSITIVE, None
      )
      self.assertSequenceEqual(sorted(got), sorted([ids[0], ids[1]]))

      # There are no negative 'hawgoo' labels.
      got = db.get_embeddings_by_label(
          'hawgoo', interface.LabelType.NEGATIVE, None
      )
      self.assertEqual(got.shape[0], 0)

    with self.subTest('get_embeddings_by_label_provenance'):
      # There is only one hawgoo labeled by a human.
      got = db.get_embeddings_by_label('hawgoo', None, 'human')
      self.assertSequenceEqual(got, [ids[0]])

      # And only one example with a 'rewbla' labeled by a machine.
      got = db.get_embeddings_by_label('rewbla', None, 'machine')
      self.assertSequenceEqual(got, [ids[0]])

    with self.subTest('count_all_labels'):
      # Finally, there are a total of three labels on ID 0.
      got = db.get_labels(ids[0])
      self.assertLen(got, 3)

    with self.subTest('get_classes'):
      got = db.get_classes()
      self.assertSequenceEqual(got, ['hawgoo', 'rewbla'])

    with self.subTest('get_class_counts'):
      # 2 positive labels for 'hawgoo' ignoring provenance, 0 for 'rewbla'.
      got = db.get_class_counts(interface.LabelType.POSITIVE)
      self.assertDictEqual(got, {'hawgoo': 2, 'rewbla': 0})

      # 1 negative label for 'rewbla', 0 for 'hawgoo'.
      got = db.get_class_counts(interface.LabelType.NEGATIVE)
      self.assertDictEqual(got, {'hawgoo': 0, 'rewbla': 1})

    with self.subTest('count_classes'):
      self.assertEqual(db.count_classes(), 2)

    with self.subTest('duplicate_labels'):
      dupe_label = interface.Label(
          ids[0], 'hawgoo', interface.LabelType.POSITIVE, 'human'
      )
      self.assertFalse(db.insert_label(dupe_label, skip_duplicates=True))
      self.assertTrue(db.insert_label(dupe_label, skip_duplicates=False))

  @parameterized.named_parameters(*DB_TYPE_NAMED_PAIRS)
  def test_brute_search_impl_agreement(self, target_db_type, source_db_type):
    rng = np.random.default_rng(42)
    source_db = test_utils.make_db(
        self.tempdir, source_db_type, 1000, rng, EMBEDDING_SIZE
    )
    target_db = test_utils.make_db(
        self.tempdir,
        target_db_type,
        1000,
        rng,
        EMBEDDING_SIZE,
        fill_random=False,
    )
    id_mapping = test_utils.clone_embeddings(source_db, target_db)

    # Check brute-force search agreement.
    query = rng.normal(size=(128,), loc=0, scale=1.0)
    results_m, _ = brutalism.brute_search(
        source_db, query, search_list_size=10, score_fn=np.dot
    )
    results_s, _ = brutalism.brute_search(
        target_db, query, search_list_size=10, score_fn=np.dot
    )
    self.assertLen(results_m.search_results, 10)
    self.assertLen(results_s.search_results, 10)
    # Search results are iterated over in sorted order.
    for r_m, r_s in zip(results_m, results_s):
      emb_m = source_db.get_embedding(r_m.embedding_id)
      emb_s = target_db.get_embedding(r_s.embedding_id)
      self.assertEqual(id_mapping[r_m.embedding_id], r_s.embedding_id)
      # TODO(tomdenton): check that the scores are the same.
      np.testing.assert_equal(emb_m, emb_s)

  def test_random_batched_iterator(self):
    rng = np.random.default_rng(42)
    db = test_utils.make_db(self.tempdir, 'in_mem', 100, rng, EMBEDDING_SIZE)
    # Test behavior when batch size divides the number of embeddings.
    with self.subTest('batch_size_divides_num_embeddings'):
      batch_size = 10
      iterator = db.random_batched_iterator(
          db.get_embedding_ids(), batch_size=batch_size, rng=rng
      )
      all_ids = set()
      for i in range(10):
        batch = next(iterator)
        self.assertLen(batch, batch_size)
        self.assertContainsSubset(batch, db.get_embedding_ids())
        all_ids.update(batch)
        self.assertLen(all_ids, min(100, (i + 1) * batch_size))

    # Test behavior when batch size does not divide the number of embeddings.
    with self.subTest('batch_size_does_not_divide_num_embeddings'):
      batch_size = 27
      iterator = db.random_batched_iterator(
          db.get_embedding_ids(), batch_size=batch_size, rng=rng
      )
      all_ids = set()
      for i in range(4):
        batch = next(iterator)
        self.assertLen(batch, batch_size)
        all_ids.update(batch)
        self.assertContainsSubset(batch, db.get_embedding_ids())
        self.assertLen(all_ids, min(100, (i + 1) * batch_size))


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2025 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for embedding audio."""

import shutil
import tempfile

from ml_collections import config_dict
from perch_hoplite.agile import embed
from perch_hoplite.agile import source_info
from perch_hoplite.agile.tests import test_utils
from perch_hoplite.db import db_loader

from absl.testing import absltest


class EmbedTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    # `self.create_tempdir()` raises an UnparsedFlagAccessError, which is why
    # we use `tempdir` directly.
    self.tempdir = tempfile.mkdtemp()

  def tearDown(self):
    super().tearDown()
    shutil.rmtree(self.tempdir)

  def test_embed_worker(self):
    classes = ['pos', 'neg']
    filenames = ['foo', 'bar', 'baz']
    test_utils.make_wav_files(self.tempdir, classes, filenames, file_len_s=6.0)

    aduio_sources = source_info.AudioSources(
        audio_globs=(
            source_info.AudioSourceConfig(
                dataset_name='test',
                base_path=self.tempdir,
                file_glob='*/*.wav',
                min_audio_len_s=0.0,
                target_sample_rate_hz=16000,
            ),
        )
    )

    in_mem_db_config = config_dict.ConfigDict()
    in_mem_db_config.embedding_dim = 32
    in_mem_db_config.max_size = 100
    db_config = db_loader.DBConfig(
        db_key='in_mem',
        db_config=in_mem_db_config,
    )

    placeholder_model_config = config_dict.ConfigDict()
    placeholder_model_config.embedding_size = 32
    placeholder_model_config.sample_rate = 16000
    model_config = embed.ModelConfig(
        model_key='placeholder_model',
        embedding_dim=32,
        model_config=placeholder_model_config,
    )

    db = db_config.load_db()

    embed_worker = embed.EmbedWorker(
        audio_sources=aduio_sources,
        model_config=model_config,
        db=db,
    )
    embed_worker.process_all()
    # The hop size is 1.0s and each file is 6.0s, so we should get 6 embeddings
    # per file. There are six files, so we should get 36 embeddings.
    self.assertEqual(db.count_embeddings(), 36)

    # Check that the metadata is set correctly.
    got_md = db.get_metadata(key=None)
    self.assertIn('audio_sources', got_md)
    self.assertIn('model_config', got_md)


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2025 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Database configuration and constructor."""

import dataclasses

from etils import epath
from ml_collections import config_dict
from perch_hoplite.db import in_mem_impl
from perch_hoplite.db import interface
from perch_hoplite.db import sqlite_usearch_impl
import tqdm


@dataclasses.dataclass
class DBConfig(interface.EmbeddingMetadata):
  """Configuration for embedding database.

  Attributes:
    db_key: Key for the database implementation to use.
    db_config: Configuration for the database implementation.
  """

  db_key: str
  db_config: config_dict.ConfigDict

  def load_db(self) -> interface.HopliteDBInterface:
    """Load the database from the specified path."""
    if self.db_key == 'sqlite_usearch':
      return sqlite_usearch_impl.SQLiteUsearchDB.create(**self.db_config)
    elif self.db_key == 'in_mem':
      return in_mem_impl.InMemoryGraphSearchDB.create(**self.db_config)
    else:
      raise ValueError(f'Unknown db_key: {self.db_key}')


def duplicate_db(
    source_db: interface.HopliteDBInterface,
    target_db_key: str,
    target_db_config: config_dict.ConfigDict,
):
  """Create a new DB and copy all data in source_db into it."""
  target_db = DBConfig(target_db_key, target_db_config).load_db()
  target_db.commit()

  # Check that the target_db is empty. If not, we'll have to do something more
  # sophisticated.
  if target_db.get_embedding_ids().shape[0] > 0:
    raise ValueError('Target DB is not empty.')

  # Clone the embeddings and get the id mapping.
  # This automatically and seamlessly clones the source info as a
  # side effect.
  emb_id_mapping = {}
  for idx in tqdm.tqdm(source_db.get_embedding_ids()):
    emb = source_db.get_embedding(idx)
    source = source_db.get_embedding_source(idx)
    target_id = target_db.insert_embedding(emb, source)
    emb_id_mapping[idx] = target_id

  target_db.commit()

  # Clone labels.
  for idx, target_idx in tqdm.tqdm(emb_id_mapping.items()):
    for lbl in source_db.get_labels(idx):
      target_label = interface.Label(
          embedding_id=target_idx,
          label=lbl.label,
          provenance=lbl.provenance,
          type=lbl.type,
      )
      target_db.insert_label(target_label)
  target_db.commit()

  # Clone the KV store, replacing the DBConfig only.
  metadata = source_db.get_metadata(key=None)
  for k, v in metadata.items():
    if k == 'db_config':
      continue
    target_db.insert_metadata(k, v)
  target_db.insert_metadata('db_config', target_db_config)
  target_db.commit()

  return target_db, emb_id_mapping


def create_new_usearch_db(
    db_path: str,
    embedding_dim: int,
) -> sqlite_usearch_impl.SQLiteUsearchDB:
  """Create a new usearch DB with the given path and embedding dimension."""
  epath.Path(db_path).parent.mkdir(parents=True, exist_ok=True)
  usearch_cfg = sqlite_usearch_impl.get_default_usearch_config(embedding_dim)
  return sqlite_usearch_impl.SQLiteUsearchDB.create(
      db_path=db_path, usearch_cfg=usearch_cfg
  )



# coding=utf-8
# Copyright 2025 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""SQLite hoplite impliementation using usearch for vector storage."""

import collections
from collections.abc import Sequence
import dataclasses
import functools
import json
import sqlite3
from typing import Any

from etils import epath
from ml_collections import config_dict
import numpy as np
from perch_hoplite.db import interface
from usearch import index as uindex


USEARCH_CONFIG_KEY = 'usearch_config'
EMBEDDINGS_TABLE = 'hoplite_embeddings'

HOPLITE_FILENAME = 'hoplite.sqlite'
UINDEX_FILENAME = 'usearch.index'

USEARCH_DTYPES = {
    'float16': uindex.ScalarKind.F16,
}


def get_default_usearch_config(
    embedding_dim: int,
) -> config_dict.ConfigDict:
  """Get a default usearch config for the given embedding dimension."""
  usearch_cfg = config_dict.ConfigDict()
  usearch_cfg.embedding_dim = embedding_dim
  usearch_cfg.dtype = 'float16'
  usearch_cfg.metric_name = 'IP'
  usearch_cfg.expansion_add = 256
  usearch_cfg.expansion_search = 128
  return usearch_cfg


@dataclasses.dataclass
class SQLiteUsearchDB(interface.HopliteDBInterface):
  """SQLite hoplite database, using USearch for vector storage.

  USearch provides both indexing for approximate nearest neighbor search and
  fast disk-based random access to vectors for the complete database. USearch
  will default to working with disk-based vectors, unless we insert or remove
  embeddings, in which case we load the index into memory and use it from there
  for all subsequent operations. On database commit, the in-memory index is
  persisted to disk.

  Attributes:
    db_path: The path to the database directory.
    db: The sqlite3 database connection.
    ui: The USearch index.
    embedding_dim: The dimension of the embeddings.
    embedding_dtype: The data type of the embeddings.
    _cursor: The sqlite3 cursor.
    _ui_loaded: Whether the USearch index was loaded in memory.
    _ui_updated: Whether the USearch index was updated since the last load and
      needs to be persisted to disk at some point in the future.
  """

  # User-provided.
  db_path: str

  # Instantiated during creation.
  db: sqlite3.Connection
  ui: uindex.Index

  # Obtained from the usearch_cfg.
  embedding_dim: int
  embedding_dtype: type[Any] = np.float16

  # Dynamic state.
  _cursor: sqlite3.Cursor | None = None
  _ui_loaded: bool = False
  _ui_updated: bool = False

  @classmethod
  def create(
      cls,
      db_path: str,
      usearch_cfg: config_dict.ConfigDict | None = None,
  ):
    db_path = epath.Path(db_path)
    db_path.mkdir(parents=True, exist_ok=True)
    hoplite_path = db_path / HOPLITE_FILENAME
    db = sqlite3.connect(hoplite_path.as_posix())
    cursor = db.cursor()
    cursor.execute('PRAGMA journal_mode=WAL;')  # Enable WAL mode
    _setup_sqlite_tables(cursor)
    db.commit()

    # TODO(tomdenton): Check that config is consistent with the DB.
    metadata = _get_metadata(cursor, None)
    if (
        USEARCH_CONFIG_KEY in metadata
        and usearch_cfg is not None
        and metadata[USEARCH_CONFIG_KEY] != usearch_cfg
    ):
      raise ValueError(
          'A usearch_cfg was provided, but one already exists in the DB.'
      )
    elif USEARCH_CONFIG_KEY in metadata:
      usearch_cfg = metadata[USEARCH_CONFIG_KEY]
    elif usearch_cfg is None:
      raise ValueError('No usearch config found in DB and none provided.')
    else:
      _insert_metadata(cursor, USEARCH_CONFIG_KEY, usearch_cfg)
      db.commit()

    usearch_dtype = USEARCH_DTYPES[usearch_cfg.dtype]
    index_path = db_path / UINDEX_FILENAME
    if index_path.exists():
      ui = uindex.Index(
          ndim=usearch_cfg.embedding_dim,
          path=index_path,
          view=True,
      )
      ui_in_memory = False
    else:
      ui = uindex.Index(
          ndim=usearch_cfg.embedding_dim,
          metric=getattr(uindex.MetricKind, usearch_cfg.metric_name),
          expansion_add=usearch_cfg.expansion_add,
          expansion_search=usearch_cfg.expansion_search,
          dtype=usearch_dtype,
          path=index_path,
          view=False,
      )
      ui_in_memory = True

    return SQLiteUsearchDB(
        db_path=db_path.as_posix(),
        db=db,
        ui=ui,
        embedding_dim=usearch_cfg.embedding_dim,
        embedding_dtype=usearch_cfg.dtype,
        _ui_loaded=ui_in_memory,
        _ui_updated=ui_in_memory,
    )

  @property
  def _sqlite_filepath(self) -> epath.Path:
    return epath.Path(self.db_path) / HOPLITE_FILENAME

  @property
  def _usearch_filepath(self) -> epath.Path:
    return epath.Path(self.db_path) / UINDEX_FILENAME

  @functools.cached_property
  def offset_dtype(self) -> type[Any]:
    """The data type of the offsets in the database."""
    if self.ui.size == 0:
      return np.float32
    # Otherwise, check the in-memory index.
    # Some old DBs may have float16 offsets, but in this case will always
    # have a single offset, so we can detect this and return float16.
    idx = self.get_one_embedding_id()
    cursor = self._get_cursor()
    cursor.execute(
        """
        SELECT he.offsets
        FROM hoplite_sources hs
        JOIN hoplite_embeddings he ON hs.id = he.source_idx
        WHERE he.id = ?;
        """,
        (int(idx),),
    )
    binary_offsets = cursor.fetchall()[0]
    if len(binary_offsets) == 2:
      return np.float16
    else:
      return np.float32

  def thread_split(self):
    """Get a new instance of the SQLite DB."""
    return self.create(self.db_path)

  def _get_cursor(self) -> sqlite3.Cursor:
    if self._cursor is None:
      self._cursor = self.db.cursor()
    return self._cursor

  def commit(self):
    self.db.commit()
    if self._cursor is not None:
      self._cursor.close()
      self._cursor = None
    if self._ui_updated:
      self.ui.save()
      self._ui_updated = False

  def get_embedding_ids(self) -> np.ndarray:
    # Note that USearch can also create a list of all keys, but it seems
    # quite slow.
    cursor = self._get_cursor()
    cursor.execute("""SELECT id FROM hoplite_embeddings;""")
    return np.array(tuple(int(c[0]) for c in cursor.fetchall()))

  def get_one_embedding_id(self) -> int:
    cursor = self._get_cursor()
    cursor.execute("""SELECT id FROM hoplite_embeddings LIMIT 1;""")
    return int(cursor.fetchone()[0])

  def insert_metadata(self, key: str, value: config_dict.ConfigDict) -> None:
    """Insert a key-value pair into the metadata table."""
    json_coded = value.to_json()
    cursor = self._get_cursor()
    cursor.execute(
        """
      INSERT INTO hoplite_metadata (key, data) VALUES (?, ?)
      ON CONFLICT (key) DO UPDATE SET data = excluded.data;
    """,
        (key, json_coded),
    )

  def get_metadata(self, key: str | None) -> config_dict.ConfigDict:
    """Get a key-value pair from the metadata table."""
    cursor = self._get_cursor()
    return _get_metadata(cursor, key)

  def get_dataset_names(self) -> tuple[str, ...]:
    """Get all dataset names in the database."""
    cursor = self._get_cursor()
    cursor.execute("""SELECT DISTINCT dataset FROM hoplite_sources;""")
    return tuple(c[0] for c in cursor.fetchall())

  def _get_source_id(
      self, dataset_name: str, source_id: str, insert: bool = False
  ) -> int | None:
    cursor = self._get_cursor()
    cursor.execute(
        """
      SELECT id FROM hoplite_sources WHERE dataset = ? AND source = ?;
    """,
        (dataset_name, source_id),
    )
    result = cursor.fetchone()
    if result is None and insert:
      cursor.execute(
          """
        INSERT INTO hoplite_sources (dataset, source) VALUES (?, ?);
      """,
          (dataset_name, source_id),
      )
      return int(cursor.lastrowid)
    elif result is None:
      return None
    return int(result[0])

  def insert_embedding(
      self, embedding: np.ndarray, source: interface.EmbeddingSource
  ) -> int:
    if embedding.shape[-1] != self.embedding_dim:
      raise ValueError('Incorrect embedding dimension.')
    cursor = self._get_cursor()
    source_id = self._get_source_id(
        source.dataset_name, source.source_id, insert=True
    )
    offset_bytes = serialize_embedding(source.offsets, self.offset_dtype)
    cursor.execute(
        """
      INSERT INTO hoplite_embeddings (source_idx, offsets) VALUES (?, ?);
    """,
        (source_id, offset_bytes),
    )
    embedding_id = int(cursor.lastrowid)

    if not self._ui_loaded:
      self.ui.load()
      self._ui_loaded = True

    self.ui.add(embedding_id, embedding.astype(self.embedding_dtype))
    self._ui_updated = True
    return embedding_id

  def count_embeddings(self) -> int:
    """Counts the number of hoplite_embeddings in the 'embeddings' table."""
    return self.ui.size

  def embedding_dimension(self) -> int:
    return self.embedding_dim

  def get_embedding(self, embedding_id: int) -> np.ndarray:
    contains = self.ui.contains(embedding_id)
    if not np.all(contains):
      raise ValueError(f'Embeddings {embedding_id} not found.')
    emb = self.ui.get(embedding_id)
    return np.array(emb)

  def get_embedding_source(
      self, embedding_id: int
  ) -> interface.EmbeddingSource:
    cursor = self._get_cursor()
    cursor.execute(
        """
        SELECT hs.dataset, hs.source, he.offsets
        FROM hoplite_sources hs
        JOIN hoplite_embeddings he ON hs.id = he.source_idx
        WHERE he.id = ?;
        """,
        (int(embedding_id),),
    )
    dataset, source, offsets = cursor.fetchall()[0]
    offsets = deserialize_embedding(offsets, self.offset_dtype)
    return interface.EmbeddingSource(dataset, str(source), offsets)

  def get_embeddings(
      self, embedding_ids: np.ndarray
  ) -> tuple[np.ndarray, np.ndarray]:
    contains = self.ui.contains(embedding_ids)
    if not np.all(contains):
      raise ValueError(f'Embeddings {embedding_ids[~contains]} not found.')
    embs = self.ui.get(embedding_ids)
    return embedding_ids, np.array(embs)

  def get_embeddings_by_source(
      self,
      dataset_name: str,
      source_id: str | None,
      offsets: np.ndarray | None = None,
  ) -> np.ndarray:
    """Get the embedding IDs for all embeddings matching the indicated source.

    Args:
      dataset_name: The name of the dataset to search.
      source_id: The ID of the source to search. If None, all sources are
        searched.
      offsets: The offsets of the source to search. If None, all offsets are
        searched.

    Returns:
      A list of embedding IDs matching the indicated source parameters.
    """
    cursor = self._get_cursor()
    if source_id is None:
      query = (
          'SELECT id, offsets FROM hoplite_embeddings '
          'WHERE hoplite_embeddings.source_idx IN ('
          '  SELECT id FROM hoplite_sources '
          '  WHERE hoplite_sources.dataset = ?'
          ');'
      )
      cursor.execute(query, (dataset_name,))
    else:
      source_idx = self._get_source_id(dataset_name, source_id, insert=False)
      query = (
          'SELECT id, offsets FROM hoplite_embeddings '
          'WHERE hoplite_embeddings.source_idx = ?;'
      )
      cursor.execute(query, (source_idx,))
    result_pairs = cursor.fetchall()
    outputs = []
    for idx, offsets_bytes in result_pairs:
      got_offsets = deserialize_embedding(offsets_bytes, self.offset_dtype)
      if offsets is not None and not np.array_equal(got_offsets, offsets):
        continue
      outputs.append(idx)
    return np.array(outputs)

  def insert_label(
      self, label: interface.Label, skip_duplicates: bool = False
  ) -> bool:
    if label.type is None:
      raise ValueError('label type must be set')
    if label.provenance is None:
      raise ValueError('label source must be set')
    if skip_duplicates and label in self.get_labels(label.embedding_id):
      return False

    cursor = self._get_cursor()
    cursor.execute(
        """
      INSERT INTO hoplite_labels (embedding_id, label, type, provenance) VALUES (?, ?, ?, ?);
    """,
        (
            int(label.embedding_id),
            label.label,
            label.type.value,
            label.provenance,
        ),
    )
    return True

  def get_embeddings_by_label(
      self,
      label: str,
      label_type: interface.LabelType | None = interface.LabelType.POSITIVE,
      provenance: str | None = None,
  ) -> np.ndarray:
    query = 'SELECT embedding_id FROM hoplite_labels WHERE label = ?'
    pred = (label,)
    if label_type is not None:
      query += ' AND type = ?'
      pred = pred + (label_type.value,)
    if provenance is not None:
      query += ' AND provenance = ?'
      pred = pred + (provenance,)
    cursor = self._get_cursor()
    results = cursor.execute(query, pred).fetchall()
    ids = np.array(tuple(int(c[0]) for c in results), np.int64)
    return np.unique(ids)

  def get_labels(self, embedding_id: int) -> tuple[interface.Label, ...]:
    cursor = self._get_cursor()
    cursor.execute(
        """
      SELECT embedding_id, label, type, provenance FROM hoplite_labels WHERE embedding_id = ?;
    """,
        (int(embedding_id),),
    )
    results = cursor.fetchall()
    return tuple(
        interface.Label(int(r[0]), r[1], interface.LabelType(r[2]), r[3])
        for r in results
    )

  def get_classes(self) -> Sequence[str]:
    cursor = self._get_cursor()
    cursor.execute('SELECT DISTINCT label FROM hoplite_labels ORDER BY label;')
    return tuple(r[0] for r in cursor.fetchall())

  def get_class_counts(
      self, label_type: interface.LabelType = interface.LabelType.POSITIVE
  ) -> dict[str, int]:
    cursor = self._get_cursor()
    # Subselect with distinct is needed to avoid double-counting the same label
    # on the same embedding because of different provenances.
    cursor.execute("""
      SELECT label, type, COUNT(*)
      FROM (
          SELECT DISTINCT embedding_id, label, type FROM hoplite_labels)
      GROUP BY label, type;
    """)
    results = collections.defaultdict(int)
    for r in cursor.fetchall():
      if r[1] == label_type.value:
        results[r[0]] = r[2]
      else:
        results[r[0]] += 0
    return results

  def print_table_values(self, table_name):
    """Prints all values from the specified table in the SQLite database."""
    cursor = self._get_cursor()
    select_query = f'SELECT * FROM {table_name};'  # Query to select all rows
    cursor.execute(select_query)

    # Fetch all rows
    rows = cursor.fetchall()

    # Print header (optional)
    if rows:
      # Column names
      print(', '.join(column[0] for column in cursor.description))

    # Print each row as a comma-separated string
    for row in rows:
      print(', '.join(str(value) for value in row))


def _setup_sqlite_tables(cursor: sqlite3.Cursor) -> None:
  """Create the needed tables in the SQLite databse.

  USearch acts as both the index and vector store. The SQLite `embeddings` table
  is used for joining embeddings with sources.

  Args:
    cursor: The SQLite cursor to use.
  """
  cursor.execute("""
  SELECT name FROM sqlite_master WHERE type='table' AND name='hoplite_labels';
  """)
  if cursor.fetchone() is not None:
    return

  # Create embedding sources table
  cursor.execute("""
      CREATE TABLE IF NOT EXISTS hoplite_sources (
          id INTEGER PRIMARY KEY AUTOINCREMENT,
          dataset STRING NOT NULL,
          source STRING NOT NULL
      );
  """)

  # Create embeddings table for joining embeddings with sources.
  cursor.execute("""
      CREATE TABLE IF NOT EXISTS hoplite_embeddings (
          id INTEGER PRIMARY KEY AUTOINCREMENT,
          source_idx INTEGER NOT NULL,
          offsets BLOB NOT NULL,
          FOREIGN KEY (source_idx) REFERENCES hoplite_sources(id)
      );
  """)

  cursor.execute("""
      CREATE TABLE IF NOT EXISTS hoplite_metadata (
          key STRING PRIMARY KEY,
          data STRING NOT NULL
      );
  """)

  # Create hoplite_labels table.
  cursor.execute("""
  CREATE TABLE IF NOT EXISTS hoplite_labels (
      id INTEGER PRIMARY KEY AUTOINCREMENT,
      embedding_id INTEGER NOT NULL,
      label STRING NOT NULL,
      type INT NOT NULL,
      provenance STRING NOT NULL,
      FOREIGN KEY (embedding_id) REFERENCES hoplite_embeddings(id)
  )""")

  # Create indices for efficient lookups.
  cursor.execute("""
  CREATE UNIQUE INDEX IF NOT EXISTS
      idx_embedding ON hoplite_embeddings (id);
  """)
  cursor.execute("""
  CREATE UNIQUE INDEX IF NOT EXISTS
      source_pairs ON hoplite_sources (dataset, source);
  """)
  cursor.execute("""
  CREATE INDEX IF NOT EXISTS embedding_source ON hoplite_embeddings (source_idx);
  """)
  cursor.execute("""
  CREATE INDEX IF NOT EXISTS idx_label ON hoplite_labels (embedding_id, label);
  """)


def _insert_metadata(
    cursor: sqlite3.Cursor, key: str, value: config_dict.ConfigDict
):
  """Insert a key-value pair into the metadata table."""
  json_coded = value.to_json()
  cursor.execute(
      """
    INSERT INTO hoplite_metadata (key, data) VALUES (?, ?)
    ON CONFLICT (key) DO UPDATE SET data = excluded.data;
  """,
      (key, json_coded),
  )


def _get_metadata(cursor: sqlite3.Cursor, key: str | None):
  """Retrieve metadata from the SQLite database."""
  if key is None:
    cursor.execute("""SELECT key, data FROM hoplite_metadata;""")
    return config_dict.ConfigDict(
        {k: json.loads(v) for k, v in cursor.fetchall()}
    )

  cursor.execute(
      """
    SELECT data FROM hoplite_metadata WHERE key = ?;
  """,
      (key,),
  )
  result = cursor.fetchone()
  if result is None:
    raise KeyError(f'Metadata key not found: {key}')
  return config_dict.ConfigDict(json.loads(result[0]))


def serialize_embedding(
    embedding: np.ndarray, embedding_dtype: type[Any]
) -> bytes:
  return embedding.astype(np.dtype(embedding_dtype).newbyteorder('<')).tobytes()


def deserialize_embedding(
    serialized_embedding: bytes, embedding_dtype: type[Any]
) -> np.ndarray:
  return np.frombuffer(
      serialized_embedding,
      dtype=np.dtype(embedding_dtype).newbyteorder('<'),
  )



# coding=utf-8
# Copyright 2025 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility functions for testing."""

from ml_collections import config_dict
import numpy as np
from perch_hoplite.db import in_mem_impl
from perch_hoplite.db import interface
from perch_hoplite.db import sqlite_usearch_impl


CLASS_LABELS = ('alpha', 'beta', 'gamma', 'delta', 'epsilon', 'zeta')


def make_db(
    path: str,
    db_type: str,
    num_embeddings: int,
    rng: np.random.Generator,
    embedding_dim: int = 128,
    fill_random: bool = True,
) -> interface.HopliteDBInterface:
  """Create a test DB of the specified type."""
  if db_type == 'in_mem':
    db = in_mem_impl.InMemoryGraphSearchDB.create(
        embedding_dim=embedding_dim,
        max_size=num_embeddings,
    )
  elif db_type == 'sqlite_usearch':
    usearch_cfg = sqlite_usearch_impl.get_default_usearch_config(embedding_dim)
    db = sqlite_usearch_impl.SQLiteUsearchDB.create(
        db_path=path, usearch_cfg=usearch_cfg
    )
  else:
    raise ValueError(f'Unknown db type: {db_type}')
  # Insert a few embeddings...
  if fill_random:
    insert_random_embeddings(db, embedding_dim, num_embeddings, rng)
  config = config_dict.ConfigDict()
  config.embedding_dim = embedding_dim
  db.insert_metadata('db_config', config)
  model_config = config_dict.ConfigDict()
  model_config.embedding_dim = embedding_dim
  model_config.model_name = 'fake_model'
  db.insert_metadata('model_config', model_config)
  db.commit()
  return db


def insert_random_embeddings(
    db: interface.HopliteDBInterface,
    emb_dim: int = 1280,
    num_embeddings: int = 1000,
    seed: int = 42,
):
  """Insert randomly generated embedding vectors into the DB."""
  rng = np.random.default_rng(seed=seed)
  np_alpha = list('ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789')
  dataset_names = ('a', 'b', 'c')
  for _ in range(num_embeddings):
    embedding = np.float32(rng.normal(size=emb_dim, loc=0, scale=0.1))
    dataset_name = rng.choice(dataset_names)
    source_name = ''.join(
        [str(a) for a in rng.choice(np_alpha, size=8, replace=False)]
    )
    offsets = rng.integers(0, 100, size=[1])
    source = interface.EmbeddingSource(dataset_name, source_name, offsets)
    db.insert_embedding(embedding, source)
  db.commit()


def clone_embeddings(
    source_db: interface.HopliteDBInterface,
    target_db: interface.HopliteDBInterface,
):
  """Copy all embeddings to target_db and provide an id mapping."""
  id_mapping = {}
  for source_id in source_db.get_embedding_ids():
    id_mapping[source_id] = target_db.insert_embedding(
        source_db.get_embedding(source_id),
        source_db.get_embedding_source(source_id),
    )
  return id_mapping


def add_random_labels(
    db: interface.HopliteDBInterface,
    rng: np.random.Generator,
    unlabeled_prob: float = 0.5,
    positive_label_prob: float = 0.5,
    provenance: str = 'test',
):
  """Insert random labels for a subset of embeddings."""
  for idx in db.get_embedding_ids():
    if rng.random() < unlabeled_prob:
      continue
    if rng.random() < positive_label_prob:
      label_type = interface.LabelType.POSITIVE
    else:
      label_type = interface.LabelType.NEGATIVE
    label = interface.Label(
        embedding_id=idx,
        label=str(rng.choice(CLASS_LABELS)),
        type=label_type,
        provenance=provenance,
    )
    db.insert_label(label)
  db.commit()



# coding=utf-8
# Copyright 2025 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base class for a searchable embeddings database."""

import abc
import dataclasses
import enum
from typing import Iterator, Sequence

from ml_collections import config_dict
import numpy as np


class LabelType(enum.Enum):
  NEGATIVE = 0
  POSITIVE = 1


@dataclasses.dataclass
class Label:
  """Label for an embedding.

  Attributes:
    embedding_id: Unique integer ID for the embedding this label applies to.
    label: Label string.
    type: Type of label (positive, negative, etc).
    provenance: Freeform field describing the annotation (eg, labeler name,
      model identifier for pseudolabels, etc).
  """

  embedding_id: int
  label: str
  type: LabelType
  provenance: str


@dataclasses.dataclass
class EmbeddingSource:
  """Source information for an embedding."""

  dataset_name: str
  source_id: str
  offsets: np.ndarray

  def __eq__(self, other):
    return (
        self.dataset_name == other.dataset_name
        and self.source_id == other.source_id
        and np.array_equal(self.offsets, other.offsets)
    )


@dataclasses.dataclass
class EmbeddingMetadata:
  """Convenience class for converting dataclasses to/from ConfigDict."""

  def to_config_dict(self) -> config_dict.ConfigDict:
    """Convert to a config dict."""
    return config_dict.ConfigDict(dataclasses.asdict(self))

  @classmethod
  def from_config_dict(
      cls, config: config_dict.ConfigDict
  ) -> 'EmbeddingMetadata':
    """Convert from a config dict."""
    return cls(**config)


class HopliteDBInterface(abc.ABC):
  """Interface for searchable embeddings database with metadata.

  The database consists of a table of embeddings with a unique id for each
  embedding, and some metadata for linking the embedding to its source.
  Additionally, a Key-Value table of ConfigDict objects is used to store
  arbitrary metadata associated with the database.

  Methods are split into 'Base' methods and 'Composite' methods. Base methods
  must be implemented for any implementation. Composite methods have a default
  implementation using the base methods, but may benefit from implementation-
  specific optimizations.
  """

  # Base methods

  @classmethod
  @abc.abstractmethod
  def create(cls, **kwargs):
    """Connect to and, if needed, initialize the database."""

  @abc.abstractmethod
  def commit(self) -> None:
    """Commit any pending transactions to the database."""

  @abc.abstractmethod
  def thread_split(self) -> 'HopliteDBInterface':
    """Get a new instance of the database with the same contents.

    For example, SQLite DB's need a distinct object in each thread.
    """

  @abc.abstractmethod
  def insert_metadata(self, key: str, value: config_dict.ConfigDict) -> None:
    """Insert a key-value pair into the metadata table."""

  @abc.abstractmethod
  def get_metadata(self, key: str | None) -> config_dict.ConfigDict:
    """Get a key-value pair from the metadata table.

    Args:
      key: String for metadata key to retrieve. If None, returns all metadata.

    Returns:
      ConfigDict containing the metadata.
    """

  @abc.abstractmethod
  def get_dataset_names(self) -> Sequence[str]:
    """Get all dataset names in the database."""

  @abc.abstractmethod
  def get_embedding_ids(self) -> np.ndarray:
    # TODO(tomdenton): Make this return an iterator, with optional shuffling.
    """Get all embedding IDs in the database."""

  @abc.abstractmethod
  def insert_embedding(
      self, embedding: np.ndarray, source: EmbeddingSource
  ) -> int:
    """Add an embedding to the database."""

  @abc.abstractmethod
  def get_embedding(self, embedding_id: int) -> np.ndarray:
    """Retrieve an embedding from the database."""

  @abc.abstractmethod
  def get_embedding_source(self, embedding_id: int) -> EmbeddingSource:
    """Get the source corresponding to the given embedding_id."""

  @abc.abstractmethod
  def get_embeddings_by_source(
      self,
      dataset_name: str,
      source_id: str | None,
      offsets: np.ndarray | None = None,
  ) -> np.ndarray:
    """Get the embedding IDs for all embeddings matching the source.

    Args:
      dataset_name: The name of the dataset to search.
      source_id: The ID of the source to search. If None, all sources are
        searched.
      offsets: The offsets of the source to search. If None, all offsets are
        searched.

    Returns:
      A list of embedding IDs matching the indicated source parameters.
    """

  @abc.abstractmethod
  def insert_label(self, label: Label, skip_duplicates: bool = False) -> bool:
    """Add a label to the db.

    Args:
      label: The label to insert.
      skip_duplicates: If True, and the label already exists, return False.
        Otherwise, the label is inserted regardless of duplicates.

    Returns:
      True if the label was inserted, False if it was a duplicate and
      skip_duplicates was True.
    Raises:
      ValueError if the label type or provenance is not set.
    """

  @abc.abstractmethod
  def embedding_dimension(self) -> int:
    """Get the embedding dimension."""

  @abc.abstractmethod
  def get_embeddings_by_label(
      self,
      label: str,
      label_type: LabelType | None = LabelType.POSITIVE,
      provenance: str | None = None,
  ) -> np.ndarray:
    """Find embeddings by label.

    Args:
      label: Label string to search for.
      label_type: Type of label to return. If None, returns all labels
        regardless of Type.
      provenance: If provided, filters to the target provenance value.

    Returns:
      An array of unique embedding id's matching the label.
    """
    # TODO(tomdenton): Allow fetching by dataset_name.

  @abc.abstractmethod
  def get_labels(self, embedding_id: int) -> Sequence[Label]:
    """Get all labels for the indicated embedding_id."""

  @abc.abstractmethod
  def get_classes(self) -> Sequence[str]:
    """Get all distinct classes (label strings) in the database."""

  @abc.abstractmethod
  def get_class_counts(
      self, label_type: LabelType = LabelType.POSITIVE
  ) -> dict[str, int]:
    """Count the number of occurences of each class in the database.

    Classes with zero matching occurences are still included in the result.

    Args:
      label_type: Type of label to count. By default, counts positive labels.
    """

  # Composite methods

  def get_one_embedding_id(self) -> int:
    """Get an arbitrary embedding id from the database."""
    return self.get_embedding_ids()[0]

  def count_embeddings(self) -> int:
    """Return a count of all embeddings in the database."""
    return len(self.get_embedding_ids())

  def count_classes(self) -> int:
    """Return a count of all distinct classes in the database."""
    return len(self.get_classes())

  def get_embeddings(
      self, embedding_ids: np.ndarray
  ) -> tuple[np.ndarray, np.ndarray]:
    """Get an array of embeddings for the indicated IDs.

    Note that the embeddings may not be returned in the same order as the
    provided embedding_id's. Thus, we suggest the usage:
    ```
    idxes, embeddings = db.get_embeddings(idxes)
    ```

    Args:
      embedding_ids: 1D array of embedding id's.

    Returns:
      Permuted array of embedding_id's and embeddings.
    """
    embeddings = [self.get_embedding(int(idx)) for idx in embedding_ids]
    return embedding_ids, np.array(embeddings)

  def get_embedding_sources(
      self, embedding_ids: np.ndarray
  ) -> tuple[EmbeddingSource, ...]:
    """Get an array of embedding sources for the indicated IDs."""
    return tuple(self.get_embedding_source(int(idx)) for idx in embedding_ids)

  def random_batched_iterator(
      self,
      ids: np.ndarray,
      batch_size: int,
      rng: np.random.RandomState,
  ) -> Iterator[np.ndarray]:
    """Yields batches embedding ids, shuffled after each of unlimited epochs."""
    if batch_size > len(ids):
      raise ValueError('Not enough ids to fill a batch.')
    rng.shuffle(ids)
    q = 0
    while True:
      if q + batch_size > len(ids):
        partial_batch = ids[q : len(ids)].copy()
        overflow = batch_size - len(partial_batch)
        rng.shuffle(ids)
        overflow_batch = ids[:overflow]
        batch = np.concatenate([partial_batch, overflow_batch], axis=0)
        q = overflow
      else:
        batch = ids[q : q + batch_size]
        q += batch_size
      yield batch.copy()



# coding=utf-8
# Copyright 2025 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functionality for embedding audio examples."""

from concurrent import futures
import dataclasses
import itertools
import threading

from absl import logging
import audioread
from ml_collections import config_dict
import numpy as np
from perch_hoplite import audio_io
from perch_hoplite.agile import source_info
from perch_hoplite.db import interface as hoplite_interface
from perch_hoplite.zoo import model_configs
from perch_hoplite.zoo import zoo_interface
import soundfile


@dataclasses.dataclass
class ModelConfig(hoplite_interface.EmbeddingMetadata):
  """Configuration for embedding model.

  Attributes:
    model_key: Key for the model wrapper class.
    embedding_dim: Dimensionality of the embedding.
    model_config: Config dict of arguments to instantiate the model wrapper.
  """

  model_key: str
  embedding_dim: int
  model_config: config_dict.ConfigDict


def worker_initializer(state):
  name = threading.current_thread().name
  state[name + 'db'] = state['db'].thread_split()


def process_source_id(
    state,
    source_id: source_info.SourceId,
):
  """Process a single audio source."""
  worker = state['worker']
  name = threading.current_thread().name
  db = state[name + 'db']
  glob = worker.audio_globs[source_id.dataset_name]
  target_sample_rate = worker.get_sample_rate_hz(source_id)
  audio_array = worker.load_audio(source_id)

  if audio_array is None:
    return
  if (
      audio_array.shape[0]
      < glob.min_audio_len_s * worker.embedding_model.sample_rate
  ):
    return

  embs = db.get_embeddings_by_source(
      dataset_name=source_id.dataset_name,
      source_id=source_id.file_id,
      offsets=np.array([source_id.offset_s], np.float16),
  )
  if embs.shape[0] > 0:
    return

  outputs = worker.embedding_model.embed(audio_array)
  embeddings = outputs.embeddings
  if embeddings is None:
    return

  emb_source_ids = []
  embs = []

  hop_size_s = worker.compute_hop_size_s(source_id, target_sample_rate)
  for t, embedding in enumerate(embeddings):
    offset_s = source_id.offset_s + t * hop_size_s
    emb_source_id = hoplite_interface.EmbeddingSource(
        dataset_name=source_id.dataset_name,
        source_id=source_id.file_id,
        offsets=np.array([offset_s], np.float16),
    )
    for channel_embedding in embedding:
      emb_source_ids.append(emb_source_id)
      embs.append(channel_embedding)
  return emb_source_ids, embs


# TODO(tomdenton): Use itertools.batched in Python 3.12+
def batched(iterable, n):
  it = iter(iterable)
  while batch := tuple(itertools.islice(it, n)):
    yield batch


class EmbedWorker:
  """Worker for embedding audio examples."""

  def __init__(
      self,
      audio_sources: source_info.AudioSources,
      model_config: ModelConfig,
      db: hoplite_interface.HopliteDBInterface,
      embedding_model: zoo_interface.EmbeddingModel | None = None,
      audio_worker_threads: int = 8,
  ):
    self.db = db
    self.model_config = model_config
    self.audio_sources = audio_sources
    self.audio_worker_threads = audio_worker_threads
    if embedding_model is None:
      model_class = model_configs.get_model_class(model_config.model_key)
      self.embedding_model = model_class.from_config(model_config.model_config)
    else:
      self.embedding_model = embedding_model
    self.audio_globs = {
        g.dataset_name: g for g in self.audio_sources.audio_globs
    }

  def _log_error(self, source_id, exception, counter_name):
    logging.warning(
        'The audio at (%s / %f) could not be loaded (%s). '
        'The exception was (%s)',
        source_id.filepath,
        source_id.offset_s,
        counter_name,
        exception,
    )

  def _update_audio_sources(self):
    """Validates the embed config and/or saves it to the DB."""
    db_metadata = self.db.get_metadata(None)
    if 'audio_sources' not in db_metadata:
      self.db.insert_metadata(
          'audio_sources', self.audio_sources.to_config_dict()
      )
      return

    db_audio_sources = source_info.AudioSources.from_config_dict(
        db_metadata['audio_sources']
    )
    merged = self.audio_sources.merge_update(db_audio_sources)
    self.db.insert_metadata('audio_sources', merged.to_config_dict())
    self.audio_sources = merged

  def _update_model_config(self):
    """Validates the model config and/or saves it to the DB."""
    db_metadata = self.db.get_metadata(None)
    if 'model_config' not in db_metadata:
      self.db.insert_metadata(
          'model_config', self.model_config.to_config_dict()
      )
      return

    db_model_config = ModelConfig(**db_metadata['model_config'])
    if self.model_config == db_model_config:
      return

    # Validate the config against the DB.
    # TODO(tomdenton): Implement compatibility checks for model configs.
    if self.model_config.model_key != db_model_config.model_key:
      raise AssertionError(
          'The configured model key does not match the model key that is '
          'already in the DB.'
      )
    if self.model_config.embedding_dim != db_model_config.embedding_dim:
      raise AssertionError(
          'The configured embedding dimension does not match the embedding '
          'dimension that is already in the DB.'
      )
    self.db.insert_metadata('model_config', self.model_config.to_config_dict())

  def update_configs(self):
    """Validates the configs and saves them to the DB."""
    self._update_model_config()
    self._update_audio_sources()
    self.db.commit()

  def get_sample_rate_hz(self, source_id: source_info.SourceId) -> int:
    """Get the sample rate of the embedding model."""
    dataset_name = source_id.dataset_name
    if dataset_name not in self.audio_globs:
      raise ValueError(f'Dataset name {dataset_name} not found in audio globs.')
    audio_glob = self.audio_globs[dataset_name]
    if audio_glob.target_sample_rate_hz == -2:
      return self.embedding_model.sample_rate
    elif audio_glob.target_sample_rate_hz == -1:
      # Uses the file's native sample rate.
      return -1
    elif audio_glob.target_sample_rate_hz > 0:
      return audio_glob.target_sample_rate_hz
    else:
      raise ValueError('Invalid target_sample_rate.')

  def load_audio(self, source_id: source_info.SourceId) -> np.ndarray | None:
    """Load audio from the indicated source and log any problems."""
    target_sample_rate_hz = self.get_sample_rate_hz(source_id)
    try:
      audio_array = audio_io.load_audio_window(
          filepath=source_id.filepath,
          offset_s=source_id.offset_s,
          sample_rate=target_sample_rate_hz,
          window_size_s=source_id.shard_len_s,
      )
      return np.array(audio_array)
    except soundfile.LibsndfileError as inst:
      self._log_error(source_id, inst, 'audio_libsndfile_error')
    except ValueError as inst:
      self._log_error(source_id, inst, 'audio_bad_offset')
    except audioread.NoBackendError as inst:
      self._log_error(source_id, inst, 'audio_no_backend')
    except EOFError as inst:
      self._log_error(source_id, inst, 'audio_eof_error')
    except RuntimeError as inst:
      if 'Soundfile is not available' in str(inst):
        self._log_error(source_id, inst, 'audio_no_soundfile')
      else:
        self._log_error(source_id, inst, 'audio_runtime_error')

  def compute_hop_size_s(
      self,
      source_id: source_info.SourceId,
      target_sample_rate_hz: int,
      model_hop_size_s: float | None = None,
  ) -> float:
    """Compute the hop size of the embedding model."""

    if model_hop_size_s is not None:
      return model_hop_size_s
    if hasattr(self.embedding_model, 'hop_size_s'):
      model_hop_size_s = float(self.embedding_model.hop_size_s)
    else:
      # TODO(tomdenton): Allow user specified hop size.
      raise ValueError('hop_size_s is not defined for the model.')
    model_sample_rate = self.embedding_model.sample_rate

    if target_sample_rate_hz == -2:
      return model_hop_size_s
    elif target_sample_rate_hz == -1:
      audio_sample_rate = source_id.sample_rate_hz
    elif target_sample_rate_hz > 0:
      audio_sample_rate = target_sample_rate_hz
    else:
      raise ValueError('Invalid target_sample_rate.')
    return model_hop_size_s * model_sample_rate / audio_sample_rate

  def embedding_exists(self, source_id: source_info.SourceId) -> bool:
    """Check whether embeddings already exist for the given source ID."""
    embs = self.db.get_embeddings_by_source(
        dataset_name=source_id.dataset_name,
        source_id=source_id.file_id,
        offsets=np.array([source_id.offset_s], np.float16),
    )
    return embs.shape[0] > 0

  def process_all(self, target_dataset_name: str | None = None, batch_size=32):
    """Process all audio examples."""
    self.update_configs()
    source_iterator = self.audio_sources.iterate_all_sources(
        target_dataset_name
    )
    state = {}
    state['db'] = self.db
    state['worker'] = self

    with futures.ThreadPoolExecutor(
        max_workers=self.audio_worker_threads,
        initializer=worker_initializer,
        initargs=(state,),
    ) as executor:
      for source_ids_batch in batched(source_iterator, batch_size):
        got = executor.map(
            process_source_id, itertools.repeat(state), source_ids_batch
        )
        # TODO(tomdenton): Consider using a db writer thread to avoid blocking.
        for result in got:
          if result is None:
            continue
          emb_source_ids, embeddings = result
          for emb_source_id, embedding in zip(emb_source_ids, embeddings):
            self.db.insert_embedding(embedding, emb_source_id)
    self.db.commit()
